VERSION 1.0 CLASS
BEGIN
  MultiUse = -1  'True
  Persistable = 0  'NotPersistable
  DataBindingBehavior = 0  'vbNone
  DataSourceBehavior  = 0  'vbNone
  MTSTransactionMode  = 0  'NotAnMTSObject
END
Attribute VB_Name = "CXMLDocument"
Attribute VB_GlobalNameSpace = False
Attribute VB_Creatable = True
Attribute VB_PredeclaredId = False
Attribute VB_Exposed = True
Option Explicit

' ***************************************************************************************
' Declarations
' ***************************************************************************************

' Interfaces
Implements IXMLReaderCallback

' Encaps
Private m_File As String                ' path of file this doc was created from
Private m_IsWellFormed As Boolean       ' syntax OK
Private m_Version As String             ' <?xml version="1.0"?>
Private m_Encoding As XMLEncodings      ' <?xml encoding="IsoLat1">
Private m_Root As CXMLNode              ' root node (only one)
Private m_CurrentParent As CXMLNode     ' for reading (temp)

' Encoding Enum
Public Enum XMLEncodings
    IsoLat1
    Utf8
    Utf16
    Windows1252
End Enum


' ***************************************************************************************
' Init/Fin
' ***************************************************************************************
Public Sub CreateNew(ByVal sRootName As String)
On Error GoTo hell
    If sRootName <> "" Then
        ' clean up
        Me.Dispose
        ' create
        Set m_Root = New CXMLNode
        If m_Root.CreateFromName(sRootName) = True Then
            m_IsWellFormed = True
            m_Version = "1.0"
            m_Encoding = XMLEncodings.IsoLat1
        Else
            Me.Dispose
        End If
    End If
    Exit Sub
hell:
End Sub
Public Sub CreateFromFile(ByVal sFile As String)
On Error GoTo hell
    ' clean up
    Me.Dispose
    ' assume success (any error will invalidate this); empty doc is ok as well
    m_IsWellFormed = True
    m_File = sFile
    ' read
    ReadFile
    Exit Sub
hell:
End Sub
Public Sub Dispose()
On Error GoTo hell
    DisposeAll m_Root
    Set m_Root = Nothing
    m_IsWellFormed = False
    m_Version = ""
    m_Encoding = XMLEncodings.IsoLat1
    m_File = ""
    Exit Sub
hell:
End Sub
Private Sub DisposeAll(ByVal node As CXMLNode)
On Error GoTo hell
    ' sub nodes
    Dim subNode As CXMLNode
    For Each subNode In node.SubNodes
        DisposeAll subNode
    Next subNode
    node.SetParent Nothing
    node.SetDocument Nothing
    node.SubNodes.Clear
    ' attribs
    Dim attrib As CXMLAttribute
    For Each attrib In node.Attributes
        attrib.SetOwnerNode Nothing
    Next attrib
    node.Attributes.Clear
    Exit Sub
hell:
End Sub


' ***************************************************************************************
' Properties
' ***************************************************************************************
Public Property Get File() As String
On Error GoTo hell
    File = m_File
    Exit Property
hell:
End Property
Public Property Get Version() As String
On Error GoTo hell
    Version = m_Version
    Exit Property
hell:
End Property
Public Property Get Encoding() As XMLEncodings
On Error GoTo hell
    Encoding = m_Encoding
    Exit Property
hell:
End Property
Public Property Let Encoding(argEncoding As XMLEncodings)
On Error GoTo hell
    m_Encoding = argEncoding
    Exit Property
hell:
End Property
Public Property Get IsWellFormed() As Boolean
On Error GoTo hell
    IsWellFormed = m_IsWellFormed
    Exit Property
hell:
End Property
Public Property Get Root() As CXMLNode
On Error GoTo hell
    Set Root = m_Root
    Exit Property
hell:
End Property


' ***************************************************************************************
' Methods
' ***************************************************************************************
Public Function SaveToFile(ByVal sFile As String) As Boolean
On Error GoTo hell
    If Not m_Root Is Nothing Then
        Dim hFile As Long: hFile = FreeFile
        If hFile <> 0 Then
            Open sFile For Output As #hFile
            If m_IsWellFormed = True Then
                Print #hFile, "<?xml version=" & Chr$(34) & m_Version & Chr$(34) & " encoding=" & Chr$(34) & GetEncodingString & Chr$(34) & "?>"
                SaveToFile = SaveRecurse(hFile, m_Root, "")
                If SaveToFile = True Then
                    m_File = sFile
                End If
            End If
        End If
    End If
    GoTo finally
hell:
finally:
On Error Resume Next
    If hFile <> 0 Then Close #hFile
End Function
Private Function SaveRecurse(ByVal hFile As Long, ByVal xn As CXMLNode, ByVal sTabs As String) As Boolean
On Error GoTo hell
    Print #hFile, sTabs & xn.ToString
    Dim xnSub As CXMLNode
    If xn.SubNodes.Count > 0 Then
        For Each xnSub In xn.SubNodes
            SaveRecurse = SaveRecurse(hFile, xnSub, sTabs & vbTab)
        Next xnSub
        Print #hFile, sTabs & "</" & xn.Name & ">"
    End If
    SaveRecurse = True
    Exit Function
hell:
End Function

' ***************************************************************************************
' Interface implementation
' ***************************************************************************************
Private Function IXMLReaderCallback_ReadTag(sTag As String) As Boolean
On Error GoTo hell
    Dim tag As CXMLTag: Set tag = New CXMLTag
    If tag.Create(sTag) = True Then
        Dim node As CXMLNode
        ' root or other?
        Select Case tag.TagType
            Case ttMeta
                Set node = New CXMLNode
                If node.CreateFromTag(tag) = True Then
                    Dim attr As CXMLAttribute
                    Set attr = node.Attributes("version")
                    If Not attr Is Nothing Then m_Version = attr.Value
                    Set attr = node.Attributes("encoding")
                    If Not attr Is Nothing Then ParseEncodingString attr.Value
                End If
            Case ttComment
                Debug.Assert False
                ' ... maybe save it or what
            Case ttTwinOpener
                If m_Root Is Nothing Then
                        Set node = New CXMLNode
                        If node.CreateFromTag(tag) = True Then
                            node.SetDocument Me
                            Set m_Root = node
                            Set m_CurrentParent = node
                        End If
                Else
                        Debug.Assert (Not m_CurrentParent Is Nothing)
                        Set node = m_CurrentParent.SubNodes.AddFromTag(tag)
                        node.SetDocument Me
                        node.SetParent m_CurrentParent
                        Set m_CurrentParent = node
                End If
            Case ttTwinCloser
                Debug.Assert (Not m_CurrentParent Is Nothing)
                Set m_CurrentParent = m_CurrentParent.Parent
            Case ttStandAlone
                Debug.Assert (Not m_CurrentParent Is Nothing)
                Set node = m_CurrentParent.SubNodes.AddFromTag(tag)
                node.SetDocument Me
                node.SetParent m_CurrentParent
        End Select
        ' continue?
        IXMLReaderCallback_ReadTag = (Not m_CurrentParent Is Nothing) Or (tag.TagType = ttMeta) Or (tag.TagType = ttComment)
    End If
    Exit Function
hell:
On Error Resume Next
    m_IsWellFormed = False
End Function


' ***************************************************************************************
' Helpers
' ***************************************************************************************
Private Function ReadFile()
On Error GoTo hell
    Set m_CurrentParent = Nothing
    Dim reader As CXMLReader
    Set reader = New CXMLReader
    reader.Create m_File
    reader.ReadAll Me
    GoTo finally
hell:
finally:
On Error Resume Next
    reader.Dipose
End Function
Private Function GetEncodingString() As String
On Error GoTo hell
    Select Case m_Encoding
        Case XMLEncodings.IsoLat1
            GetEncodingString = "ISO-8859-1"
        Case XMLEncodings.Utf8
            GetEncodingString = "UTF-8"
        Case XMLEncodings.Utf16
            GetEncodingString = "UTF-16"
        Case XMLEncodings.Windows1252
            GetEncodingString = "Windows-1252"
        Case Else
            GetEncodingString = ""
    End Select
    Exit Function
hell:
End Function
Private Function ParseEncodingString(ByVal sEncStr As String) As XMLEncodings
On Error GoTo hell
    ' supposedly not casesense
    If StrComp(sEncStr, "ISO-8859-1", vbTextCompare) = 0 Then
        m_Encoding = XMLEncodings.IsoLat1
    ElseIf StrComp(sEncStr, "UTF-8", vbTextCompare) = 0 Then
        m_Encoding = XMLEncodings.Utf8
    ElseIf StrComp(sEncStr, "UTF-16", vbTextCompare) = 0 Then
        m_Encoding = XMLEncodings.Utf16
    ElseIf StrComp(sEncStr, "Windows-1252", vbTextCompare) = 0 Then
        m_Encoding = XMLEncodings.Windows1252
    Else
        m_Encoding = XMLEncodings.IsoLat1
    End If
    Exit Function
hell:
End Function


